import cv2
from aiortc import VideoStreamTrack
from av import VideoFrame
import fractions

class CameraStream(VideoStreamTrack):
    def __init__(self):
        super().__init__()
        self.cap = cv2.VideoCapture(0)
        self.frame_count = 0
        self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    async def recv(self):
        self.frame_count += 1
        ret, frame = self.cap.read()
        
        if not ret:
            print("Failed to read frame from camera")
            return None
        
        # Create and return an AIORTC VideoFrame
        video_frame = VideoFrame.from_ndarray(frame)
        video_frame.pts = self.frame_count
        video_frame.time_base = fractions.Fraction(1, 30)
        return video_frame

    def __del__(self):
        """Release camera and video writer when the object is destroyed."""
        if hasattr(self, 'cap') and self.cap.isOpened():
            self.cap.release()


import asyncio
import json
from aiortc import RTCDataChannel

class DataStream:
    def __init__(self, message=None, interval: float = 1.0):
        self.channel = None
        self._running = False
        self.message = message or {"type": "message", "message": "Hello!"}
        self.interval = interval

    def set_channel(self, channel: RTCDataChannel):
        self.channel = channel

        @channel.on("open")
        def on_open():
            print("Data channel is open!")
            self._running = True
            asyncio.create_task(self._send_data())

        @channel.on("close")
        def on_close():
            print("Data channel is closed")
            self._running = False

        @channel.on("message")
        def on_message(msg):
            try:
                data = json.loads(msg)
                print("Received JSON:", data)
            except json.JSONDecodeError:
                print("Received non-JSON message:", msg)

    async def _send_data(self):
        while self._running and self.channel and self.channel.readyState == "open":
            # Inject timestamp into message
            message_data = self.message.copy() if isinstance(self.message, dict) else {"message": self.message}

            self.channel.send(json.dumps(message_data))
            await asyncio.sleep(self.interval)

    def set_message(self, message):
        """
        Set the message to be sent.
        Accepts either a string or a dict (for JSON support).
        """
        if isinstance(message, (str, dict)):
            self.message = message
        else:
            raise ValueError("Message must be a string or dictionary.")

    def set_interval(self, interval: float):
        self.interval = interval



from aiortc import RTCPeerConnection, RTCSessionDescription
from aiortc.contrib.signaling import TcpSocketSignaling

class WebRTCServer:
    def __init__(self, ip_address, port):
        self.ip_address = ip_address
        self.port = port
        self.camera_stream = None
        self.data_streams = []

    def add_camera_stream(self, stream):
        """Add a media track to the server."""
        self.camera_stream = stream

    def add_data_stream(self, data_stream):
        """Add a data stream to the server."""
        self.data_streams.append(data_stream)

    async def _setup_peer_connection(self):
        pc = RTCPeerConnection()

        # Add camera stream
        pc.addTrack(self.camera_stream)

        # Create a single data channel and link to data streams
        if self.data_streams:
            channel = pc.createDataChannel("chat")
            for ds in self.data_streams:
                ds.set_channel(channel)

        return pc

    async def _handle_signaling(self, pc, signaling):
        """Handle signaling messages for one session."""
        await signaling.connect()

        @pc.on("connectionstatechange")
        async def on_connection_state_change():
            print(f"Connection state changed: {pc.connectionState}")
            if pc.connectionState in ["failed", "disconnected", "closed"]:
                print("Cleaning up connection...")
                await pc.close()

        offer = await pc.createOffer()
        await pc.setLocalDescription(offer)

        try:
            await signaling.send(pc.localDescription)
        except Exception as e:
            print(f"Failed to send offer: {e}")
            return

        while True:
            obj = await signaling.receive()
            if isinstance(obj, RTCSessionDescription):
                await pc.setRemoteDescription(obj)
                print("Remote description set")
            elif obj is None:
                print("Signaling ended")
                break

    async def handle_connection(self):
        """Continuously listen for and handle WebRTC connections."""
        while True:
            print(f"WebRTC listening on {self.ip_address}:{self.port}")
            signaling = TcpSocketSignaling(self.ip_address, self.port)
            pc = await self._setup_peer_connection()

            try:
                await self._handle_signaling(pc, signaling)
            except Exception as e:
                print(f"Error during signaling: {e}")
            finally:
                await pc.close()
                try:
                    await signaling.close()
                except Exception:
                    pass
                print("Connection closed, waiting for new client...")
                
    async def run(self):
        if not self.camera_stream:
            raise RuntimeError("WebRTCServer requires at least one media track.")
        await self.handle_connection()



from aiortc import RTCPeerConnection
from aiortc.contrib.signaling import TcpSocketSignaling

class WebRTCClient:
    def __init__(self, ip, port):
        self.pc = RTCPeerConnection()
        self.signaling = TcpSocketSignaling(ip, port)
        self.video_track = None
        self.data_channel = None

        @self.pc.on("track")
        def on_track(track):
            print(f"Received track: {track.kind}")
            if track.kind == "video":
                self.video_track = track

        @self.pc.on("datachannel")
        def on_datachannel(channel):
            print(f"Data channel received: {channel.label}")
            self.data_channel = channel

            @channel.on("message")
            def on_message(message):
                print(f"Received data: {message}")

        @self.pc.on("connectionstatechange")
        async def on_connection_state_change():
            print(f"Connection state changed: {self.pc.connectionState}")
            if self.pc.connectionState in ["failed", "disconnected", "closed"]:
                print("Cleaning up connection...")
                await self.close()
                await self.connect()

    async def answer(self):
        offer = await self.signaling.receive()
        await self.pc.setRemoteDescription(offer)
        answer = await self.pc.createAnswer()
        await self.pc.setLocalDescription(answer)
        await self.signaling.send(self.pc.localDescription)

    async def get_frame(self):
        """Returns a single frame as ndarray (BGR) from incoming video track."""
        if self.video_track is None:
            return None

        frame = await self.video_track.recv()
        return frame.to_ndarray(format="bgr24")

    async def connect(self):
        await self.signaling.connect()
        await self.answer()

    async def close(self):
        await self.signaling.close()
        await self.pc.close()
